"""
Result class represents the result of a client call.
"""


class Result:
    """
    Result contains the following attributes:
    - message: a message string containing the description of the result
    - response: a RubyAem::Response response from AEM
    - data: the data payload, which can be of any type depending on the API call
    e.g. is_* and exists method provide result with boolean data.
    Some API calls result doesn't contain any data.
    """

    def __init__(self, message, response, data=None):
        """
        Initialise a result.

        :param message: result message
        :param response: HTTP response
        :return PythonAEM Result instance
        """
        self.message = message
        self.response = response
        self.data = data



"""
Swagger class contains logic related to swaggeraem.
"""

import re


def operation_to_method(operation):
    """
    Convert pythonaem spec's operation (consistent with Swagger spec's operationId)
    into swaggeraem's generated method name.

    :param operation: operation ID
    :return swaggeraem method name
    """
    return re.sub(r'(?<=[a-z])(?=[A-Z])', '_', operation).lower()



"""
PythonAEM error, contains a message and PythonAEM Result object
"""


class Error(RuntimeError):
    """
    PythonAEM error, contains a message and PythonAEM Result object
    useful for debugging the result and response when an error occurs
    """

    def __init__(self, message, result):
        """
        Initialise a result.

        :param message: result message
        :param resi;t: PythonAEM Result
        :return PythonAEM Error instance
        """
        super().__init__()
        self.message = message
        self.result = result



"""
Response handlers for no payload.
"""
import json
from ..error import Error
from ..result import Result


class SimpleHandler:
    """
    Response handlers for no payload.
    """

    def __init__(self, response, response_spec, call_params):
        self.response = response
        self.response_spec = response_spec
        self.call_params = call_params

    def simple(self):
        """
        Simple handler by returning result that contains status and message as
        configured in conf/spec.yaml AS-IS.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        response_message = self.response_spec['message'].format(**self.call_params)
        return Result(response_message, self.response)

    def simple_true(self):
        """
        Simple handler with boolean true result data.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        result = self.simple()
        result.data = True
        return result

    def simple_false(self):
        """
        Simple handler with boolean false result data.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        result = self.simple()
        result.data = False
        return result

    def simple_nil(self):
        """
        Simple handler with nil result data.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return RubyAem::Result
        """
        result = self.simple()
        result.data = None
        return result

    def simple_error(self):
        """
        Simple handler with raised error.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return RubyAem::Result
        """
        response_message = self.response_spec['message'].format(**self.call_params)
        result = self.simple()
        print(response_message)
        raise Error(response_message, result)

    def simple_body(self):
        """
        Simple handler with response body as result data.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        result = self.simple()
        result.data = self.response.body
        return result

    def json_aem_health_check(self):
        """
        Handle AEM Health Check Servlet JSON payload.

        :param response: HTTP response containing status_code, body, and headers
        :param response_spec: response specification as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        print(">>>>>>>")
        print(type(self.response))
        print(self.response.body)
        print(type(self.response.body))
        print("bb")
        _json = json.loads(self.response.body)
        print("aa")
        print(_json)
        response_message = self.response_spec['message'].format(**self.call_params)
        print(_json)
        data = _json['results']
        return Result(response_message, self.response, data)



"""
PythonAEM is a Python API client for interacting with AEM API.
"""
import pkgutil
import swaggeraem
from swaggeraem.apis.custom_api import CustomApi
from swaggeraem.apis.sling_api import SlingApi
import yaml
from .client import Client
from .resources.aem import Aem
from .resources.flush_agent import FlushAgent
from .resources.replication_agent import ReplicationAgent


class PythonAem:
    """
    Aem class represents an AEM client instance.
    """

    def __init__(self, conf):
        """
        Initialise a PythonAEM instance
        :param conf: <dict>
        Configuration dictionary of the following configuration values:
        - username: username used to authenticate to AEM instance, default: 'admin'
        - password: password used to authenticate to AEM instance, default: 'admin'
        - protocol: AEM instance protocol (http or https), default: 'http'
        - host: AEM instance host name, default: 'localhost'
        - port: AEM instance port, default: 4502
        """
        conf = sanitise_conf(conf)

        configuration = swaggeraem.Configuration(
            username=conf['username'],
            password=conf['password'],
            host=f'{conf["protocol"]}://{conf["host"]}:{conf["port"]}'
        )

        apis = {}
        with swaggeraem.ApiClient(configuration) as api_client:
            apis['custom'] = CustomApi(api_client)
            apis['sling'] = SlingApi(api_client)

        spec_stream = pkgutil.get_data(__name__, "conf/spec.yaml")
        spec = yaml.safe_load(spec_stream)

        self.client = Client(apis, spec)

    def aem(self):
        """
        Create an AEM instance.
        :return: new PythonAEM Resources AEM instance
        """
        return Aem(self.client)

    def flush_agent(self, run_mode, name):
        """
        Create a flush agent instance.
        :param run_mode: AEM run mode (author or publish)
        :param name: the flush agent's name, e.g. some-flush-agent
        :return: new PythonAEM Resources FlushAgent instance
        """
        return FlushAgent(self.client, run_mode, name)

    def replication_agent(self, run_mode, name):
        """
        Create a replication agent instance.
        :param run_mode: AEM run mode (author or publish)
        :param name: the replication agent's name, e.g. some-replication-agent
        :return: new PythonAEM Resources ReplicationAgent instance
        """
        return ReplicationAgent(self.client, run_mode, name)


def sanitise_conf(conf):
    """
    Set default configuration values
    :param conf: <dict>
    :return:
    """
    conf['username'] = conf.get('username', 'admin')
    conf['password'] = conf.get('password', 'admin')
    conf['protocol'] = conf.get('protocol', 'http')
    conf['host'] = conf.get('host', 'localhost')
    conf['port'] = conf.get('port', 4502)
    return conf



"""
AEM resources.
"""


class Aem:
    """
    AEM class contains API calls related to managing the AEM instance itself.
    """

    def __init__(self, client):
        """
        Initialise an AEM agent.
        :param client: PythonAEM Client
        :return new PythonAEM Resources AEM instance
        """
        self.client = client
        self.call_params = {
        }

    def get_aem_health_check(self, opts):
        """
        Retrieve AEM Health Check.
        This is a custom API and requires
        https://github.com/shinesolutions/aem-healthcheck
        to be installed.

        :param opts: optional parameters:
        - tags: comma separated tags of AEM Health Check tags
        - combine_tags_or: if true, the check needs to only pass
          one of the check tags in order to get the health check pass,
          if false, all check tags need to pass in order to get the health check pass.
        :return PythonAEM Result
        """
        self.call_params = {**self.call_params, **opts}
        return self.client.call(self.__class__.__name__, 'get_aem_health_check', self.call_params)



"""
Response wraps HTTP response data returned by swaggeraem.
"""


class Response:
    """
    Response wraps HTTP response data returned by swaggeraem.
    """

    def __init__(self, status_code, body, headers):
        """
        Initialise a result.

        :param status_code: HTTP status code
        :param body: HTTP response body
        :param headers: HTTP headers
        :return PythonAEM Response instance
        """
        self.status_code = status_code
        self.body = body
        self.headers = headers



"""
FlushAgent class contains API calls related to managing an AEM flush agent.
"""

from urllib.parse import urlparse


class FlushAgent:
    """
    FlushAgent class contains API calls related to managing an AEM flush agent.
    """

    def __init__(self, client, run_mode, name):
        """
        Initialise a flush agent.
        :param client: PythonAEM Client
        :param run_mode: AEM run mode: author or publish
        :param name: the flush agent's name, e.g. some-flush-agent
        :return new PythonAEM Resources FlushAgent instance
        """
        self.client = client
        self.call_params = {
            'run_mode': run_mode,
            'name': name
        }

    def create_update(
        self,
        title,
        description,
        dest_base_url,
        opts={
          'log_level': 'error',
          'retry_delay': 30_000
          }):
        """
        Create or update a flush agent.

        :param title: flush agent title
        :param description: flush agent description
        :param dest_base_url: base URL of the agent target destination,
            e.g. http://somedispatcher:8080
        :param opts: optional parameters:
        - log_level: error, info, debug, default is error
        - retry_delay: in milliseconds, default is 30_000
        :return PythonAEM Result
        """
        self.call_params['title'] = title
        self.call_params['description'] = description
        self.call_params['dest_base_url'] = dest_base_url

        uri = urlparse(dest_base_url)
        self.call_params['ssl'] = 'relaxed' if uri.scheme == 'https' else ''

        self.call_params = {**self.call_params, **opts}

        return self.client.call(self.__class__.__name__, 'create_update', self.call_params)

    def delete(self):
        """
        Delete the flush agent.

        :return PythonAEM Result
        """
        return self.client.call(self.__class__.__name__, 'delete', self.call_params)

    def disable(self):
        """
        Disable the flush agent.

        :return PythonAEM Result
        """
        return self.client.call(self.__class__.__name__, 'disable', self.call_params)

    def enable(self):
        """
        Enable the flush agent.

        :return PythonAEM Result
        """
        return self.client.call(self.__class__.__name__, 'enable', self.call_params)

    def exists(self):
        """
        Check whether the flush agent exists or not.
        If the flush agent  exists, this method returns a true result data,
        false otherwise.

        :return PythonAEM Result
        """
        return self.client.call(self.__class__.__name__, 'exists', self.call_params)



"""
Client class makes Swagger AEM API calls and handles the response as
configured in conf/spec.yaml .
"""

from swaggeraem.exceptions import ApiException
from .error import Error
from .handlers.simple import SimpleHandler
from .response import Response
from .result import Result
from .swagger import operation_to_method


class Client:
    """
    Client class makes Swagger AEM API calls and handles the response as
    configured in conf/spec.yaml .
    """

    def __init__(self, apis, spec):
        """
        Initialise a client.
        :param apis: a dictionary of Swagger AEM client's API instances
        :param spec: pythonaem specification
        :return new Client instance
        """
        self.apis = apis
        self.spec = spec

    def call(self, clazz, action, call_params):
        """
        Make an API call using the relevant Swagger AEM API client.
        Clazz and action parameters are used to identify the action, API, and params
        from pythonaem specification, alongside the response handlers.
        Call parameters are used to construct HTTP request parameters based on the
        specification.

        :param clazz: the class name of the caller resource
        :param action: the action of the API call
        :param call_params: API call parameters
        :return PythonAEM Result
        """
        resource_name = clazz.lower()
        resource = self.spec[resource_name]
        action_spec = resource['actions'][action]

        api = self.apis[str(action_spec['api'].lower())]
        operation = action_spec['operation']

        params = []

        if 'required' in action_spec['params']:
            required_params = action_spec['params']['required']
            for key, value in required_params.items():
                params.append(value.format(**call_params))

        keyword_params = {}
        if 'optional' in action_spec['params']:
            optional_params = action_spec['params']['optional']
            if isinstance(optional_params, dict):
                for key, value in optional_params.items():
                    self.__add_optional_param(key, value, keyword_params, call_params)

            elif isinstance(optional_params, list):
                for key in optional_params:
                    self.__add_optional_param(key, None, keyword_params, call_params)

        responses_spec = {}
        if 'responses' in resource:
            responses_spec.update(resource['responses'])

        if 'responses' in action_spec:
            responses_spec.update(action_spec['responses'])

        try:
            swagger_method = operation_to_method(operation)
            swagger_http_method_name = '{}_with_http_info'.format(swagger_method)
            data, status_code, headers = getattr(
                api, swagger_http_method_name
            )(*params, **keyword_params)
            response = Response(status_code, data, headers)
        except ApiException as exception:
            response = Response(exception.status, exception.body, exception.headers)

        return self.handle(response, responses_spec, call_params)()

    # pylint: disable=no-else-return
    def handle(self, response, responses_spec, call_params):
        """
        Handle a response based on status code and a given list of response specifications.
        If none of the response specifications contains the status code, a failure result
        will then be returned.

        :param response: response containing HTTP status code, body, and headers
        :param responses_spec: a list of response specifications as configured in conf/spec.yaml
        :param call_params: API call parameters
        :return PythonAEM Result
        :raise PythonAEM Error when the response status code is unexpected
        """
        if response.status_code in responses_spec:
            response_spec = responses_spec[response.status_code]
            handler = response_spec['handler']
            handler_class = SimpleHandler(response, response_spec, call_params)
            return getattr(handler_class, handler)
        else:
            message = "Unexpected response\nstatus code: {}\nheaders: {}\nbody: {}".format(
                response.status_code, response.headers, response.body
            )
            result = Result(message, response)
            raise Error(message, result)

    # Switch boolean type  check to isinstance after integration and unit tests have better coverage
    # pylint: disable=unidiomatic-typecheck
    def __add_optional_param(self, key, value, params, call_params):
        # if there is no value in optional param spec,
        # then only add optional param that is set in call parameters
        if value is None:
            if key in call_params:
                params[str(key)] = call_params[str(key)]
        # if value is provided in optional param spec,
        # then apply variable interpolation the same way as required param
        elif isinstance(value, str):
            params[str(key)] = value.format(**call_params)
        elif type(value) == bool:
            params[str(key)] = str(value).lower().format(**call_params)
        else:
            params[str(key)] = value



import os
import setuptools
from setuptools import sic
import yaml

readme_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md")
with open(readme_file, "r") as readme_fh:
    readme = readme_fh.read()

info_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "conf/info.yaml")
with open(info_file, "r") as info_fh:
    info = yaml.load(info_fh, Loader=yaml.FullLoader)

setuptools.setup(
    name="pythonaem",
    version=sic(info["version"]),
    author="Shine Solutions",
    author_email="opensource@shinesolutions.com",
    keywords=["Adobe Experience Manager (AEM) API"],
    description="Python client for AEM API",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/shinesolutions/pythonaem",
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pyyaml==5.4.1",
        "swaggeraem==1.2.0"
    ],
)
